#!/bin/bash

error() {
	echo "$@" >&2
	exit 1
}

help() {
	cat <<EOF
Usage: $0 [options]

Image arguments:
  --boot file: imx-boot image file
  --image file: disk image to write to mmc

One time programming arguments:
  --mac hwaddr1 hwaddr2: both mac addresses to write to fuses
  --mmcpart: toggle to configure gp partitions and enhanced mmc
  --sn board_id: serial number to write to gp0/gp1 fuse,
                 treated as a string e.g. 00C700010008 will be
		 yakushima-es1 (00C7), Lot 1, SN x0008

Behavioural arguments:
  --wait: wait for usb device forever
  --reset: reset after flashing image (dev mode)
  --monitor|-m usbpath: install update on specified device, passed as is to uuu -m
                        (usbpath is the first column of uuu -lsusb, e.g. '-m 1:3')
                        (unlike uuu, only the last occurence is kept)
  --verbose: show uuu file before executing
  --dry-run: do not run uuu (implies verbose)
  --uuu: path to uuu

Special arguments:
  --help: this help

EOF
}

parse_args() {
	local OK

	while (($# > 0)); do
		case "$1" in
		-h|--help)
			help
			exit 0
			;;

		--boot|--uboot)
			(( $# >= 2 )) || error "Not enough arguments for $1"
			[[ -r "$2" ]] || error "Can not read $2"
			UBOOT=$(readlink -e "$2") || error "cannot resolve path to $2"
			OK=1
			shift 2
			;;
		--image)
			(( $# >= 2 )) || error "Not enough arguments for $1"
			[[ -r "$2" ]] || error "Can not read $2"
			IMAGE=$(readlink -e "$2") || error "cannot resolve path to $2"
			OK=1
			shift 2
			;;
		--mac)
			(( $# >= 3 )) || error "$1 requires two arguments"
			MAC1="$2"
			MAC2="$3"
			OK=1
			shift 3
			;;
		--mmcpart)
			MMCPART=1
			OK=1
			shift
			;;
		--sn)
			(( $# >= 2 )) || error "$1 requires an argument"
			SN="$2"
			OK=1
			shift 2
			;;
		--wait)
			WAIT=1
			shift
			;;
		-m|--monitor)
			(( $# >= 2 )) || error "$1 requires an argument"
			MONITOR="$2"
			shift 2
			;;
		--dry-run)
			DRYRUN=1
			VERBOSE=1
			shift
			;;
		--reset)
			POWEROFF=
			shift
			;;
		--verbose)
			VERBOSE=1
			shift
			;;
		--uuu)
			(( $# >= 2 )) || error "$1 requires an argument"
			UUU="$2"
			[[ -x "$UUU" ]] || error "$UUU is not executable"
			shift 2
			;;
		*)
			error "Unknown arg: $1"
			;;
		esac
	done

	[[ -n "$OK" ]] ||
		error "You probably want to run this with an image file or program something? see --help for options"
}

prepare_src() {
	# uuu will prefix absolute paths with wherever the conf file is,
	# so we can't really use it for tmp... just make a link to these.
	if [[ -n "$UBOOT" ]]; then
		ln -s "$UBOOT" "$CONF_DIR/uboot"
		UBOOT="uboot"
	fi

	if [[ -n "$IMAGE" ]]; then
		# Automatically use the uboot embedded in image if available
		if [[ -z "$UBOOT" ]]; then
			dd if="$IMAGE" of="$CONF_DIR/uboot" bs=1k skip=32 count=4096 status=none
			if strings "$CONF_DIR/uboot" | grep -q U-Boot; then
				UBOOT=uboot
			fi
		fi
		ln -s "$IMAGE" "$CONF_DIR/image"
		IMAGE="image"
	fi
}

gen_uuu_conf_boot() {
	cat <<EOF
uuu_version 1.2.39

# Note: this script is based on emmc_all
# + fuse programming + mmc hwpartition + final boot

EOF
	if [[ -n "$UBOOT" ]]; then
		cat <<EOF
# This command will be run when i.MX6/7 i.MX8MM, i.MX8MQ
SDP: boot -f $UBOOT

# This command will be run when ROM support stream mode
# i.MX8QXP, i.MX8QM
SDPS: boot -f $UBOOT

# These commands will be run when use SPL and will be skipped if no spl
# SDPU will be deprecated. please use SDPV instead of SDPU
# {
SDPU: delay 1000
SDPU: write -f $UBOOT -offset 0x57c00
SDPU: jump
# }

# These commands will be run when use SPL and will be skipped if no spl
# if (SPL support SDPV)
# {
SDPV: delay 1000
SDPV: write -f $UBOOT -skipspl
SDPV: jump
# }

EOF
	fi
}

gen_uuu_conf() {
	gen_uuu_conf_boot

	if [[ -n "$MAC1" && -n "$MAC2" ]]; then
		local FUSE0 FUSE1 FUSE2
		MAC1="${MAC1//:/}"
		MAC2="${MAC2//:/}"
		[[ "${#MAC1}" = 12 ]] || error "Mac $MAC1 not the expected length"
		[[ "${#MAC2}" = 12 ]] || error "Mac $MAC2 not the expected length"
		[[ "$MAC1" != "$MAC2" ]] || error "Mac 1 and 2 are equal!"
		FUSE0="${MAC1:4:8}"
		FUSE1="${MAC2:8:4}${MAC1:0:4}"
		FUSE2="${MAC2:0:8}"

		cat <<EOF
# burn mac addresses
FB: ucmd fuse prog -y 9 0 $FUSE0 $FUSE1 $FUSE2
# lock macs
FB: ucmd fuse prog -y 0 0 4000

EOF
		TWO_STEPS=1
	fi
	if [[ -n "$MMCPART" ]]; then
		[[ -n "$UBOOT" ]] || [[ -n "$IMAGE" ]] || echo "WARNING: Will require a reinstall after flashing"
		cat <<EOF
# partition mmc
FB: ucmd mmc dev 2
# 0x3ab for old MMC, 4f0 for new one (-4 for GP)
# check with 'mmc extcsd read /dev/mmcblk2 | grep ENH_SIZE'
FB: ucmd mmc hwpartition complete user enh 0 $(((0x4f0-4)*16384)) wrrel on gp1 16384 enh gp2 16384 enh gp3 16384 enh gp4 16384 enh
# also enable RST (RST_n_ENABLE=1)
FB: ucmd mmc rst-function 2 1

EOF
		TWO_STEPS=1
	fi

	if [[ -n "$SN" ]]; then
		local FUSE0 FUSE1
		# Sanity check
		[[ "${#SN}" = 12 ]] || error "Expecting a 12 characters long string"
		FUSE0=${SN:0:4}
		FUSE1=${SN:4:8}
		[[ -n "${FUSE0//[0-9a-fA-F]}" ]] && error "Product part of serial number contains non-hex character: $FUSE0"
		[[ -n "${FUSE1//[0-9]}" ]] && error "SN part of serial number contains non-decimal character: $FUSE1"
		cat <<EOF
# burn serial number
FB: ucmd fuse prog -y 14 0 $FUSE0 $FUSE1
# lock serial number
FB: ucmd fuse prog -y 0 0 100000

EOF
		TWO_STEPS=1
	fi

	if [[ -n "$TWO_STEPS" ]]; then
		if [[ -n "$UBOOT" ]]; then
			cat <<EOF
# reset to apply fuse/part changes
FB: acmd reset
FB: done
EOF
			exec > "$CONF_DIR/script2.uuu"
			gen_uuu_conf_boot
		elif [[ -n "$IMAGE" ]]; then
			error "cannot do an operation that needs reset (writing fuse or mmcpart) and write image without uboot"
		fi
	fi

	# only flash uboot if image was also present or if no
	# fuse/emmc writing was required (two_steps is empty)
	if [[ -n "$IMAGE" ]] || { [[ -n "$UBOOT" ]] && [[ -z "$TWO_STEPS" ]]; }; then
		cat <<'EOF'
FB: ucmd setenv fastboot_dev mmc
FB: ucmd setenv mmcdev ${emmc_dev}
FB: ucmd mmc dev ${emmc_dev}
EOF
		[[ -n "$IMAGE" ]] && echo "FB: flash -raw2sparse all $IMAGE"
		[[ -n "$UBOOT" ]] && echo "FB: flash bootloader $UBOOT"

		cat <<'EOF'
FB: ucmd if env exists emmc_ack; then ; else setenv emmc_ack 0; fi;
FB: ucmd mmc partconf ${emmc_dev} ${emmc_ack} 1 0
EOF
	fi
	if [[ -n "$IMAGE" ]]; then
		cat <<EOF
FB: ucmd setenv mmcroot /dev/mmcblk2p1 rootwait ro ${POWEROFF:+firstboot_poweroff}
# we need usb disabled when linux boots with computer
# plugged in.
# es1 dtb should work for both boards for flashing purposes,
# revisit later if required
FB: ucmd setenv fdt_overlays armadillo_iotg_g4-nousb-overlay.dtb
FB: acmd boot
EOF
	elif [[ -z "$POWEROFF" ]]; then
		echo "FB: acmd reset"
	else
		echo "# This turns the board off (sets TOP+DP_EN to SNVS_LPCR)"
		echo "FB: acmd mw 0x30370038 60 1"
	fi
	echo "FB: done"
}

main() {
	local UUU=uuu CONF_DIR UBOOT IMAGE
	local MAC1 MAC2 MMCPART SN SN_PREFIX
	local VERBOSE DRYRUN TWO_STEPS="" POWEROFF=1 WAIT MONITOR

	parse_args "$@"

	if [[ -z "$WAIT" ]] && [[ -z "$DRYRUN" ]]; then
		if [[ $("$UUU" -lsusb | awk '$1 ~ /^'"${MONITOR:-[0-9].*}"'$/ { print "ok" }') = "" ]]; then
			error "No uuu-compatible device detected${MONITOR:+ on port $MONITOR}, please check device is plugged in and powered on"
		fi
	fi

	CONF_DIR=$(mktemp -t -d flash_image.XXXXXX) \
		|| error "Could not create temporary directory"
	trap "rm -rf '$CONF_DIR'" EXIT

	prepare_src
	gen_uuu_conf > "$CONF_DIR/script.uuu"
	if [[ -n "$VERBOSE" ]]; then
		cat "$CONF_DIR/script.uuu"
		[[ -e "$CONF_DIR/script2.uuu" ]] && cat "$CONF_DIR/script2.uuu"
	fi
	if [[ -z "$DRYRUN" ]]; then
		declare -a UUU_ARGS=( )
		[[ -n "$MONITOR" ]] && UUU_ARGS+=( -m "$MONITOR" )
		[[ -z "$WAIT" ]] && UUU_ARGS+=( -t 5 -T 5 )
		"$UUU" "${UUU_ARGS[@]}" "$CONF_DIR/script.uuu" || error "uuu failed"
		if [[ -e "$CONF_DIR/script2.uuu" ]]; then
			"$UUU" "${UUU_ARGS[@]}" "$CONF_DIR/script2.uuu" || error "uuu failed"
		fi
	fi
}

main "$@"
